/*
 * Copyright (C) 2005-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
/** \file filter.c the stand-alone filter program */

#include "para.h"

#include "filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "filter.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "error.h"

/** init the array of errors for para_filter */
INIT_FILTER_ERRLISTS;

/** the task that reads from stdin */
static struct stdin_task stdin_task_struct;
/** pointer to the stdin task */
static struct stdin_task *sit = &stdin_task_struct;

/** the task that filters the data */
static struct filter_chain filter_chain_struct;
/** pointer to the filter chain */
static struct filter_chain *fc = &filter_chain_struct;

/** the task that writes converted data to stdout */
static struct stdout_task stdout_task_struct;
/** pointer to the stdout task */
static struct stdout_task *sot = &stdout_task_struct;

/** gengetopt struct that holds the command line args */
static struct filter_args_info conf;

INIT_STDERR_LOGGING(conf.loglevel_arg);

static void filter_event_handler(struct task *t)
{
	PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-t->ret));
	unregister_task(t);
}

static void open_filters(void)
{
	struct filter_node *fn;

	list_for_each_entry(fn, &fc->filters, node) {
		fn->filter->open(fn);
		PARA_INFO_LOG("opened %s filter\n", fn->filter->name);
		fc->outbuf = fn->buf;
		fc->out_loaded = &fn->loaded;
	}
}

static int init_filter_chain(void)
{
	int i, filter_num;
	struct filter_node *fn;

	INIT_LIST_HEAD(&fc->filters);

	fc->inbuf = sit->buf;
	fc->in_loaded = &sit->loaded;
	fc->input_eof = &sit->eof;
	fc->eof = 0;
	fc->output_eof = &sot->eof;
	fc->task.private_data = fc;
	fc->task.pre_select = filter_pre_select;
	fc->task.event_handler = filter_event_handler;
	sprintf(fc->task.status, "filter chain");

	for (i = 0; i < conf.filter_given; i++) {
		char *fa = conf.filter_arg[i];
		fn = para_calloc(sizeof(struct filter_node));
		filter_num = check_filter_arg(fa, &fn->conf);
		if (filter_num < 0) {
			free(fn);
			return filter_num;
		}
		fn->fc = fc;
		INIT_LIST_HEAD(&fn->callbacks);
		fn->filter = &filters[filter_num];
		PARA_DEBUG_LOG("adding %s to filter chain\n", fn->filter->name);
		list_add_tail(&fn->node, &fc->filters);
	}
	if (list_empty(&fc->filters))
		return -E_NO_FILTERS;
	open_filters();
	return 1;
}

static int parse_config(int argc, char *argv[])
{
	static char *cf; /* config file */
	struct stat statbuf;
	int i;

	if (filter_cmdline_parser(argc, argv, &conf))
		return -E_FILTER_SYNTAX;
	HANDLE_VERSION_FLAG("filter", conf);
	if (!cf) {
		char *home = para_homedir();
		cf = make_message("%s/.paraslash/filter.conf", home);
		free(home);
	}
	if (!stat(cf, &statbuf)) {
		if (filter_cmdline_parser_configfile(cf, &conf, 0, 0, 0))
			return -E_FILTER_SYNTAX;
	}
	if (!conf.list_filters_given)
		return 1;
	printf("available filters: ");
	for (i = 0; filters[i].name; i++)
		printf("%s%s%s", i? " " : "", filters[i].name,
			filters[i].parse_config? "*": "");
	printf("\nFilters marked with \"*\" have further command line options. Try\n"
		"\tpara_filter -f '<filtername> -h'\nfor more information.\n");
	exit(EXIT_SUCCESS);
}

/**
 * para_filter's main function.
 *
 * para_filter reads data from stdin, converts it by using a chain
 * of filters (specified on the command line) and writes the resulting
 * data to stdout.
 *
 * \param argc number of command line options
 * \param argv vector of arguments
 *
 * \return \a EXIT_SUCCESS on success, EXIT_FAILURE on errors.
 */
int main(int argc, char *argv[])
{
	int ret;
	struct sched s;

	stdin_set_defaults(sit);
	sit->buf = para_malloc(sit->bufsize),

	filter_init(filters);
	ret = parse_config(argc, argv);
	if (ret < 0)
		goto out;
	ret = init_filter_chain();
	if (ret < 0)
		goto out;

	stdout_set_defaults(sot);
	sot->buf = fc->outbuf;
	sot->loaded = fc->out_loaded;
	sot->input_eof = &fc->eof;

	register_task(&sot->task);
	register_task(&fc->task);
	register_task(&sit->task);
	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	ret = sched(&s);
out:
	free(sit->buf);
	close_filters(fc);
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", PARA_STRERROR(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
