/*
 * Copyright (C) 2004-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file string.c memory allocation and string handling functions */

#include "para.h"
#include "string.h"

#include <sys/time.h> /* gettimeofday */
#include <pwd.h>
#include <sys/utsname.h> /* uname() */
#include <string.h>

#include "error.h"

/**
 * paraslash's version of realloc()
 *
 * \param p pointer to the memory block, may be NULL
 * \param size desired new size
 *
 * A wrapper for realloc(3). It calls \p exit(\p EXIT_FAILURE) on errors,
 * i.e. there is no need to check the return value in the caller.
 *
 * \return A pointer to  the newly allocated memory, which is suitably aligned
 * for any kind of variable and may be different from \p p.
 *
 * \sa realloc(3)
 */
__must_check __malloc void *para_realloc(void *p, size_t size)
{
	/*
	 * No need to check for NULL pointers: If p is NULL, the  call
	 * to realloc is equivalent to malloc(size)
	 */
	if (!(p = realloc(p, size))) {
		PARA_EMERG_LOG("%s", "realloc failed, aborting\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * paraslash's version of malloc()
 *
 * \param size desired new size
 *
 * A wrapper for malloc(3) which exits on errors.
 *
 * \return A pointer to the allocated memory, which is suitably aligned for any
 * kind  of variable.
 *
 * \sa malloc(3)
 */
__must_check __malloc void *para_malloc(size_t size)
{
	void *p = malloc(size);

	if (!p) {
		PARA_EMERG_LOG("%s", "malloc failed, aborting\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * paraslash's version of calloc()
 *
 * \param size desired new size
 *
 * A wrapper for calloc(3) which exits on errors.
 *
 * \return A pointer to the allocated and zeroed-out memory, which is suitably
 * aligned for any kind  of variable.
 *
 * \sa calloc(3)
 */
__must_check __malloc void *para_calloc(size_t size)
{
	void *ret = para_malloc(size);

	memset(ret, 0, size);
	return ret;
}

/**
 * paraslash's version of strdup()
 *
 * \param s string to be duplicated
 *
 * A wrapper for strdup(3). It calls \p exit(EXIT_FAILURE) on errors, i.e.
 * there is no need to check the return value in the caller.
 *
 * \return A pointer to the duplicated string. If \p s was the NULL pointer,
 * an pointer to an empty string is returned.
 *
 * \sa strdup(3)
 */
__must_check __malloc char *para_strdup(const char *s)
{
	char *ret;

	if ((ret = strdup(s? s: "")))
		return ret;
	PARA_EMERG_LOG("%s", "strdup failed, aborting\n");
	exit(EXIT_FAILURE);
}

/**
 * allocate a sufficiently large string and print into it
 *
 * \param fmt usual format string
 *
 * Produce output according to \p fmt. No artificial bound on the length of the
 * resulting string is imposed.
 *
 * \return This function either returns a pointer to a string that must be
 * freed by the caller or aborts without returning.
 *
 * \sa printf(3)
 */
__must_check __printf_1_2 __malloc char *make_message(const char *fmt, ...)
{
	char *msg;

	PARA_VSPRINTF(fmt, msg);
	return msg;
}

/**
 * paraslash's version of strcat()
 *
 * \param a string to be appended to
 * \param b string to append
 *
 * Append \p b to \p a.
 *
 * \return If \p a is NULL, return a pointer to a copy of \p b, i.e.
 * para_strcat(NULL, b) is equivalent to para_strdup(b). If \p b is NULL,
 * return \p a without making a copy of \p a.  Otherwise, construct the
 * concatenation \p c, free \p a (but not \p b) and return \p c.
 *
 * \sa strcat(3)
 */
__must_check __malloc char *para_strcat(char *a, const char *b)
{
	char *tmp;

	if (!a)
		return para_strdup(b);
	if (!b)
		return a;
	tmp = make_message("%s%s", a, b);
	free(a);
	return tmp;
}

/**
 * paraslash's version of dirname()
 *
 * \param name pointer to the full path
 *
 * Compute the directory component of \p name
 *
 * \return If \p name is \ NULL or the empty string, return \p NULL.
 * Otherwise, Make a copy of \p name and return its directory component. Caller
 * is responsible to free the result.
 */
__must_check __malloc char *para_dirname(const char *name)
{
	char *p, *ret;

	if (!name || !*name)
		return NULL;
	ret = para_strdup(name);
	p = strrchr(ret, '/');
	if (!p)
		*ret = '\0';
	else
		*p = '\0';
	return ret;
}

/**
 * paraslash's version of basename()
 *
 * \param name Pointer to the full path
 *
 * Compute the filename component of \p name
 *
 * \return If \p name is \p NULL or the empty string, return \p NULL,
 * Otherwise, make a copy of \p name and return its filename component. Caller
 * is responsible to free the result.
 */
__must_check __malloc char *para_basename(const char *name)
{
	char *p;

	if (!name || !*name)
		return NULL;
	p = strrchr(name, '/');
	if (!p)
		return para_strdup(name);
	p++;
	if (!*p)
		return NULL;
	return para_strdup(p);
}

/**
 * cut trailing newline
 *
 * \param buf the string to be chopped.
 *
 * Replace the last character in \p buf by zero if it is euqal to
 * the newline character.
 */
void chop(char *buf)
{
	int n = strlen(buf);
	if (!n)
		return;
	if (buf[n - 1] == '\n')
		buf[n - 1] = '\0';
}

/**
 * get a random filename
 *
 * This is by no means a secure way to create temporary files in a hostile
 * direcory like \p /tmp. However, it is OK to use for temp files, fifos,
 * sockets that are created in ~/.paraslash. Result must be freed by the
 * caller.
 *
 * \return a pointer to a random filename.
 */
__must_check __malloc char *para_tmpname(void)
{
	struct timeval now;
	gettimeofday(&now, NULL);
	srand(now.tv_usec);
	return make_message("%08i", rand());
}

/**
 * create unique temporary file
 *
 * \param template the template to be passed to mkstemp()
 * \param mode the desired mode of the tempfile
 *
 * This wrapper for mkstemp additionally uses fchmod() to
 * set the given mode of the tempfile if mkstemp() returned success.
 *
 * \return The file descriptor of the temp file just created on success.
 * On errors, -E_MKSTEMP or -E_FCHMOD is returned.
 */
__must_check int para_mkstemp(char *template, mode_t mode)
{
	int tmp, fd = mkstemp(template);

	if (fd < 0)
		return -E_MKSTEMP;
	tmp = fchmod(fd, mode);
	if (tmp >= 0)
		return fd;
	close(fd);
	unlink(template);
	return -E_FCHMOD;
}

/**
 * get the logname of the current user
 *
 * \return A dynammically allocated string that must be freed by the caller. On
 * errors, the string "unknown user" is returned, i.e. this function never
 * returns NULL.
 *
 * \sa getpwuid(3)
 */
__must_check __malloc char *para_logname(void)
{
	struct passwd *pw = getpwuid(getuid());
	return para_strdup(pw? pw->pw_name : "unknown_user");
}

/**
 * get the home directory of the current user
 *
 * \return A dynammically allocated string that must be freed by the caller. If
 * the home directory could not be found, this function returns "/tmp".
 */
__must_check __malloc char *para_homedir(void)
{
	struct passwd *pw = getpwuid(getuid());
	return para_strdup(pw? pw->pw_dir : "/tmp");
}

/**
 * split string and return pointers to its parts.
 *
 * \param args the string to be split
 * \param argv_ptr  pointer to the list of substrings
 * \param delim delimiter
 *
 * This function modifies \a args by replacing each occurance of \a delim by
 * zero. A \p NULL-terminated array of pointers to char* is allocated dynamically
 * and these pointers are initialized to point to the broken-up substrings
 * within \a args. A pointer to this array is returned via \a argv_ptr. It's OK
 * to call this function with \a args \a == \p NULL.
 *
 * \return The number of substrings found in \a args.
 */
__must_check unsigned split_args(char *args, char ***argv_ptr, const char *delim)
{
	char *p = args;
	char **argv;
	size_t n = 0, i, j;

	p = args + strspn(args, delim);
	for (;;) {
		i = strcspn(p, delim);
		if (!i)
			break;
		p += i;
		n++;
		p += strspn(p, delim);
	}
	*argv_ptr = para_malloc((n + 1) * sizeof(char *));
	argv = *argv_ptr;
	i = 0;
	p = args + strspn(args, delim);
	while (p) {
		argv[i] = p;
		j = strcspn(p, delim);
		if (!j)
			break;
		p += strcspn(p, delim);
		if (*p) {
			*p = '\0';
			p++;
			p += strspn(p, delim);
		}
		i++;
	}
	argv[n] = NULL;
	return n;
}

/**
 * ensure that file descriptors 0, 1, and 2 are valid
 *
 * Common approach that opens /dev/null until it gets a file descriptor greater
 * than two.
 *
 * \sa okir's Black Hats Manual.
 */
void valid_fd_012(void)
{
	while (1) {
	int     fd;

		fd = open("/dev/null", O_RDWR);
		if (fd < 0)
			exit(EXIT_FAILURE);
		if (fd > 2) {
			close(fd);
			break;
		}
	}
}

/**
 * get the own hostname
 *
 * \return A dynammically allocated string containing the hostname.
 *
 * \sa uname(2)
 */
__malloc char *para_hostname(void)
{
	struct utsname u;

	uname(&u);
	return para_strdup(u.nodename);
}
